#include <cmath>
#include "color.h"
#include "colorcurve.h"
#include "mesh.h"
#include "renderer.h"
#include "shader.h"
#include "texture2d.h"

using namespace std;

namespace Msp {
namespace GL {

ColorCurve::ColorCurve():
	shprog("colorcurve.glsl"),
	quad(get_fullscreen_quad())
{
	shdata.uniform("source", 0);
	shdata.uniform("curve", 1);

	curve.storage(LUMINANCE, 256);
	curve.set_min_filter(LINEAR);
	curve.set_wrap(CLAMP_TO_EDGE);
	texturing.attach(1, curve);

	set_exposure_adjust(0.0f);
	set_brightness_response(0.4f);
	set_linear();
}

void ColorCurve::set_exposure_adjust(float e)
{
	shdata.uniform("exposure", pow(2.0f, e));
}

void ColorCurve::set_brightness_response(float b)
{
	if(b<=0 || b>1)
		throw invalid_argument("ColorCurve::set_brightness_response");
	float t = (b<1 ? pow(b, 1/(1-b)) : 0.0f);
	shdata.uniform("brightness_response", b, t, pow(t, b));
}

void ColorCurve::set_peak(float)
{
}

void ColorCurve::set_brightness(float b)
{
	set_brightness_response(1/b);
}

void ColorCurve::set_gamma(float g)
{
	if(g<0.1 || g>10)
		throw invalid_argument("ColorCurve::set_gamma");

	unsigned char curve_data[256];
	for(unsigned i=0; i<256; ++i)
		curve_data[i] = pow(i/255.0f, 1/g)*255+0.5f;
	curve.image(0, LUMINANCE, UNSIGNED_BYTE, curve_data);
}

void ColorCurve::set_srgb()
{
	unsigned char curve_data[256];
	curve_data[0] = 0;
	for(unsigned i=1; i<256; ++i)
		curve_data[i] = to_srgb(i/255.0f)*255+0.5f;
	curve.image(0, LUMINANCE, UNSIGNED_BYTE, curve_data);
}

void ColorCurve::set_linear()
{
	unsigned char curve_data[256];
	for(unsigned i=0; i<256; ++i)
		curve_data[i] = i;
	curve.image(0, LUMINANCE, UNSIGNED_BYTE, curve_data);
}

void ColorCurve::render(Renderer &renderer, const Texture2D &color_buf, const Texture2D &)
{
	texturing.attach(0, color_buf);

	Renderer::Push push(renderer);
	renderer.set_shader_program(&shprog, &shdata);
	renderer.set_texturing(&texturing);
	quad.draw(renderer);
}

} // namespace GL
} // namespace Msp
